/* 
 * This file is main.c, part of the Skipjack Coordinate Decoder
 *
 * Most of the Skipjack Coordinate Decoder is free software: you can
 * redistribute it and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Most of the Skipjack Coordinate Decoder is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Skipjack Coordinate Decoder.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Martin Oldfield <m@mjo.tc>
 *
 * Version 0.1 : 2012-12-31, Initial release
 *
 * Note: The project also includes the skip32 algorithm written by
 *       Greg Rose, which says the following about licenses and
 *       copyright:
 *            "Not copyright, no rights reserved."
 *
 * For more information about this project see 
 *     http://mjo.tc/atelier/2012/12/coord-decoder.html
 */

#include <xc.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

#include "skip32.h"

__CONFIG(FOSC_INTRC_NOCLKOUT & WDTE_OFF & PWRTE_OFF & MCLRE_OFF & CP_OFF & CPD_OFF & BOREN_OFF & IESO_OFF & FCMEN_OFF & LVP_OFF & DEBUG_OFF);

static void display_swirl(void);
static void display_msg(const uint8_t msg[]);
static void display_coord(const int32_t x, uint8_t pos, uint8_t neg);
static void disp_decimal(int offset, uint32_t x, int max_width, bool zeropad);
static void disp_hex(int offset, uint32_t x, int max_width, bool zeropad);
static void display_tick(void);
static void read_encoder(void);
static void read_encoder_button(void);
static void read_button(void);
static void inc_crypt(int8_t sense);


// These definitions match the schematic e.g. RC6 is connected
// to the 'a' segment
#define seg_a (1 << 6)
#define seg_b (1 << 4)
#define seg_c (1 << 2)
#define seg_d (1 << 0)
#define seg_e (1 << 7)
#define seg_f (1 << 5)
#define seg_g (1 << 3)
#define seg_p (1 << 1)

// Hex font for 7-seg display
static const uint8_t font[16] = { 
  seg_a + seg_b + seg_c + seg_d + seg_e + seg_f,
          seg_b + seg_c,
  seg_a + seg_b +         seg_d + seg_e +         seg_g,
  seg_a + seg_b + seg_c + seg_d +                 seg_g,
          seg_b + seg_c +                 seg_f + seg_g,
  seg_a +         seg_c + seg_d +         seg_f + seg_g,
  seg_a +         seg_c + seg_d + seg_e + seg_f + seg_g,
  seg_a + seg_b + seg_c,
  seg_a + seg_b + seg_c + seg_d + seg_e + seg_f + seg_g,
  seg_a + seg_b + seg_c + seg_d +         seg_f + seg_g,
  seg_a + seg_b + seg_c +         seg_e + seg_f + seg_g,
                  seg_c + seg_d + seg_e + seg_f + seg_g,
  seg_a +                 seg_d + seg_e + seg_f,
          seg_b + seg_c + seg_d + seg_e +         seg_g,
  seg_a +                 seg_d + seg_e + seg_f + seg_g,
  seg_a +                         seg_e + seg_f + seg_g,
};

// some letters in the font: W is obviously a fudge---
// there's only enough width for a single U!
const uint8_t font_n = seg_a | seg_b | seg_c | seg_e | seg_f;
const uint8_t font_s = seg_a | seg_f | seg_g | seg_c | seg_d;
const uint8_t font_e = seg_a | seg_g | seg_d | seg_e | seg_f;
const uint8_t font_w = seg_b | seg_c | seg_d | seg_e | seg_f;

#define font_H (seg_b | seg_c | seg_e | seg_f | seg_g)
#define font_E (seg_a | seg_g | seg_d | seg_e | seg_f)
#define font_L (seg_d | seg_e | seg_f)
#define font_O (seg_a | seg_b | seg_c | seg_d | seg_e | seg_f)

// These 8 bytes are essentially our frame buffer: application
// code can write here, and the display will automatically
// reflect the changes courtesy of a natty interrupt routine.
static volatile uint8_t display[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

// System clock, which ticks when TMR2 overflows (~2kHz)
static volatile uint16_t t = 0;

static volatile uint8_t  selected_digit = 7;

static volatile bool     update_disp;
static volatile bool     encoder_enabled;
static volatile uint8_t  disp_mode      = 5;
static volatile uint8_t  disp_mode_lock = 0;
static volatile uint16_t t_transition  = 0;
static volatile bool     timed_transition = false;

static int32_t n_offset;
static int32_t e_offset;

static volatile uint32_t plain   = 0;

// This is the initial value of the cyphertext
static volatile uint32_t crypt   = 0x12345678;
 
// This structure is read from the flash chip if fitted
// and configures the decoding algorithm!
// 
// The offsets are specified in milli-minutes so so e.g. N 51 -> 51 * 60000 = 3060000 = 0x2eb120
static uint8_t inbuff[26] = { 
  // Start up message: remember that the LSB is on the right `so'
  // this is backwards
  seg_p, seg_p, seg_p, font_O, font_L, font_L, font_E, font_H,
  // North offset: N 51 12.345 => 0x002ee159
  0x59, 0xe1, 0x2e, 0x00,
  // East offset:  E  0  9.876 => 0x00002694
  0x94, 0x26, 0x00, 0x00,
  // Skipjack32 Key
  0x00, 0x99, 0x88, 0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11,
};

static uint8_t *hello_msg = inbuff;
static uint8_t *skip_key  = inbuff + 16;

// Set this to non-zero to scan the flash on startup
#define SCAN_FLASH (1)

#if SCAN_FLASH
static bool scan_flash(void);
#endif

////////////////////////////////////////////////////////////////////////
//
// UNI/O library (alpha quality only)
//
// This code is rather time sensitive and it seemed easier to
// make it work if I used macros instead of functions...
//
// The basic idea is to synchronize all the transitions with timer 2
// and we assume that we can use it as we wish! We also assume that
// RB4 is the UNI/O bus.
//
// TMR2:
//   freq = 500kHz / 8 = 62.5kHz	
//   period = 16us
//
// Basically we put a call to TMR2_WAIT before everything which is
// time-critical.
//
// There might still be some jitter but that seems better addressed in
// assembler, and this seems to work.
//
// Note:
//   The UNIO_SEND macros take four TMR2 ticks to sent a bit
//   giving a bit rate of 64us / 16kHz c.f. 10--100kHz in spec
//
/////////

// Low level routines

// wait until timer2 rolls over
#define TMR2_WAIT         { while(!TMR2IF) { ; } TMR2IF = 0; }

// Send a bit
#define UNIO_SEND(x)      {                \
                            TMR2_WAIT;	   \
                            RB4 = (1-(x)); \
			    TRISB4 = 0;    \
			    TMR2_WAIT;     \
                            TMR2_WAIT;     \
			    RB4 = (x);     \
			    TMR2_WAIT;     \
                          }

// Read back the SAK pulse
#define UNIO_READ_SAK(x)  {                 \
                            TMR2_WAIT;      \
                            TRISB4 = 1;     \
                            TMR2_WAIT;      \
                            x = (RB4 << 1); \
                            TMR2_WAIT;      \
                            TMR2_WAIT;      \
                            x |= RB4;       \
                          }

// High level, user callable routines
#define UNIO_INIT {                                                   \
                    PR2 = 7;					      \
                  						      \
		    /* Stay high for 30 x 16us ~ 0.5ms */             \
                    TRISB4 = 0;					      \
                    RB4 = 1;					      \
                    for(uint8_t i = 0; i < 30; i++)		      \
                      TMR2_WAIT;				      \
                  						      \
                    /* Release from POR: low pulse for about 16us */  \
                    RB4 = 0;					      \
                    TMR2_WAIT;					      \
                    RB4 = 1;					      \
                  						      \
                    /* standby : 60 x 16us ~ 1ms c.f. 600us min */    \
                    for(uint8_t i = 0; i < 60; i++)		      \
                      TMR2_WAIT;				      \
                  						      \
                    /* Set the bus ready for future fun */   	      \
                    RB4 = 0;					      \
                    TMR2_WAIT;					      \
                  };


// Send byte x, with MAK = a, and set stat = SAK status
#define UNIO_TX(x,a,stat) {                                       \
                            for(uint8_t m = 128; m != 0; m >>= 1) \
			      { UNIO_SEND((x) & m ? 1 : 0); };    \
			    UNIO_SEND(a);                         \
			    UNIO_READ_SAK(stat);                  \
                          }

// Read byte x, with MAK = a, and set stat = SAK status
#define UNIO_RX(x,a,stat) {                                       \
                            x = 0;                                \
			    for(uint8_t i = 0; i < 8; i++)        \
			      {                                   \
				TMR2_WAIT;                        \
                                TMR2_WAIT;                        \
                                TMR2_WAIT;                        \
                                x <<= 1;                          \
                                TMR2_WAIT;                        \
                                x |= RB4;                         \
			      }                                   \
                            UNIO_SEND(a);                         \
			    UNIO_READ_SAK(stat);                  \
                           }


////////////////////////////////////////////////////////////////////////

void main(void)
{
  OSCCONbits.IRFC = 0x07; // 8MHz clock

  // Port B: 0-3  input with weak pullups
  //         4    SCIO to UNI/O, WPU
  TRISB  = 0xff;
  ANSELH = 0;
  nRBPU  = 0;
  WPUB   = 0x1f;     // RB0-4 should have weak pull-ups
  PORTB  = 0;
  
  TRISA = 0xff;     // We'll set single bits low as we scan the LED array
  PORTA = 0xff;
  TRISC = 0x00;     // ...but these are always outputs

  // Freq of TMR2 is 500kHz: 2MHz instruction clock, devide-by-4 prescaler
  T2CONbits.T2CKPS = 1;
  T2CONbits.TMR2ON = 1;

#if SCAN_FLASH
  // change the last digit of the message to show whether we read the flash
  hello_msg[0] = scan_flash() ? font[0xf] : font_n;    
#endif

  n_offset = *((int32_t *)(inbuff + 8));
  e_offset = *((int32_t *)(inbuff + 12));

  PR2    = 0xff;
  PEIE   = 1;
  TMR2IE = 1;           // ...interrupt on TMR2 overflow (500kHz / 256 ~ 2kHz )

  ei();                // enable interrupts

  while(1)
    {
      update_disp = false;

      uint8_t old_disp_mode = disp_mode;
      switch(disp_mode)
	{
	case 0:
	  encoder_enabled = true;
	  disp_hex(0, crypt, 8, true);
	  display[selected_digit] |= seg_p;
	  break;

	case 1:
	  encoder_enabled = false;
	  disp_mode_lock = 1;

	  for(uint8_t i = 0; i < 8; i++)
	    display[i] = 0x00;

	  plain = skip32(skip_key, crypt, 0);

	  display_swirl();

	  disp_mode++;
	  disp_mode_lock = 0;
	  break;

	case 2:
	  display_coord(n_offset + (plain >> 16L), font_n, font_s);
	  break;

	case 3:
	  display_coord(e_offset + (plain & 0xffffL), font_e, font_w);
	  break;
	  
	case 5:
	  display_msg(hello_msg);
	  t_transition = t + 5 * 2000;  // about 5s
	  timed_transition = true;
	  break;

	case 4:
	case 6:
	default:
	  disp_mode = 0;
	  timed_transition = false;
	  break;
	}

      if (disp_mode != old_disp_mode)
	update_disp = true;

      while(!update_disp)
	;
    }
}

#if SCAN_FLASH
// Scan UNI/O flash memory and write data to inbuff
// returning true if we succeeded, false otherwise
//
// This isn't really production quality code: we ought to put a
// checksum on the flash data for a start! We might also want
// to fall-back to inbuilt coords if the read fails midway...
// 
static bool scan_flash(void)
{
  uint8_t sak; 

  UNIO_INIT;
   
  // startup
  UNIO_TX(0x55, 1, sak);   // Can regard start header as sending 0x55
  // don't expect SAK here
  
  UNIO_TX(0xa0, 1, sak);   // 0xa0 is the device address on the UNI/O bus
  if (sak != 1)            // check SAK
    return false;
  
  UNIO_TX(0x03, 1, sak);   // 0x03 => read
  UNIO_TX(0x00, 1, sak);   // high byte of address
  UNIO_TX(0x00, 1, sak);   // low byte of address
  
  // read config data into inbuff
  uint8_t u2;
  for(uint8_t i = 0; i < sizeof(inbuff); i++)
    {
      UNIO_RX(u2,1,sak);
      inbuff[i] = u2;
    }

  // read and discard an extra byte so we can easily end with noMAK
  UNIO_RX(u2,0,sak);

  return true;
}
#endif

// Display some pretty nonsense
static void display_swirl(void)
{
  const uint8_t f[8] = { seg_a, seg_b, seg_c, seg_d, seg_e, seg_f, seg_g, 0 };

  for(uint8_t i = 0; i < 8; i++)
    {
      for(uint8_t j = 0; f[j] != 0; j++)
	{
	  display[i] = f[j];
	  for(unsigned int k = 0; k < 5000; k++)
	    ;
	}
    }
}

static void display_msg(const uint8_t msg[])
{
  for(int i = 0; i < 8; i++)
    display[i] = msg[i];
}

static void display_coord(const int32_t x, uint8_t pos, uint8_t neg)
{
  for(int i = 0; i < 8; i++)
    display[i] = 0;
  
  display[7] = (x >=0) ? pos : neg;

  uint32_t y = (x < 0) ? -x : x;

  uint32_t min = y % 60000L;
  uint32_t deg = (y - min) / 60000L;

  disp_decimal(0, min, 5, true);
  disp_decimal(5, deg, 3, false);

  display[3] |= seg_p;
  display[5] |= seg_p;
}

static void disp_decimal(int offset, uint32_t x, int max_width, bool zeropad)
{
  for(int i = 0; i < max_width; i++)
    {
      int d = x % 10;
      display[offset + i] = font[d];
      x = (x - d) / 10;
      
      if (x == 0 && !zeropad)
	break;
    }
}

static void disp_hex(int offset, uint32_t x, int max_width, bool zeropad)
{
  for(int i = 0; i < max_width; i++)
    {
      int d = x & 0xf;
      display[offset + i] = font[d];
      x >>= 4;
      
      if (x == 0 && !zeropad)
	break;
    }
}
     
void interrupt handler(void)
{
  if (!(TMR2IE && TMR2IF))
    return;

  // timer2 fires at about 2kHz
  TMR2IF = 0;

  display_tick();
  read_button();
  
  if (encoder_enabled)
    {
      read_encoder_button();
      read_encoder();
    }
  
  if (timed_transition && t == t_transition)
    {
      disp_mode++;
      update_disp = true;
      timed_transition = false;
    }

  t++;

  return;
}

// update the LED display from global state variables
//  - designed to be called from e.g. timer interrupt 
static void display_tick(void)
{
  static uint8_t digit_mask = 0x80;
  static uint8_t digit      = 0;

  uint8_t n = display[digit];
  
  PORTC = n ^ 0xff;
  TRISA = digit_mask ^ 0xff;

  if (digit == 7)
    {
      digit = 0;
      digit_mask = 0x80;
    }
  else
    {
      digit++;
      digit_mask >>= 1;
    }
}

// update state from button state
//  - increment selected_digit for every press
//  - crude debouncing inherent in auto-repeat
static void read_encoder_button(void)
{
  static uint8_t  locked = 0;
  static uint16_t t_lockout;

  if (locked)
    {
      if (t > t_lockout)
	locked = 0;
      else
	return;
    }
      
  if (PORTBbits.RB2)
    return;

  // ignore other changes for 200 ticks
  // => if button is held down we get auto-repeat
  locked = 1;
  t_lockout = t + 800;

  selected_digit = (selected_digit - 1) & 7;
  update_disp = true;

}

// update state from encoder
static void read_encoder(void)
{
  const int8_t tx_map[16] = { 0,-1,1,0,
			      1,0,0,-1,
			     -1,0,0,1,
			      0,1,-1,0 };

  static uint8_t old_state = 0x03;
  uint8_t state = PORTB & 0x03;

  int8_t tx = -tx_map[(old_state << 2) | state];
  old_state = state;

  if (tx == 0)
    return;

  // We get multiple transitions for each 'click' so 
  // keep a fractional click count
  const  uint8_t   dx = 64;
  static uint8_t   x  = 0;  
  if (tx < 0)
    {
      if (x < dx)
	inc_crypt(-1);
      x -= dx;
    }
  else
    {
      if (x >= 255 - (dx - 1))
	inc_crypt(+1);
      x += dx;
    }
}
  
static void inc_crypt(int8_t sense)
{
  uint8_t  k = selected_digit << 2;
  uint32_t crypt_inc  = 0x1L << k;
  uint32_t new_crypt  = (sense < 0) ? crypt - crypt_inc : crypt + crypt_inc;
  uint32_t new_mask = 0xfL << k;
  uint32_t old_mask = new_mask ^ 0xffffffff;

  crypt = (crypt & old_mask) | (new_crypt & new_mask);
  update_disp = true;
}

static void read_button(void)
{
  static uint8_t  locked = 0;
  static uint16_t t_lockout;

  if (locked)
    {
      if (t > t_lockout)
	locked = 0;
      else
	return;
    }
      
  if (PORTBbits.RB3)
    return;

  // ignore other changes for 200 ticks
  // => if button is held down we get auto-repeat
  locked = 1;
  t_lockout = t + 400;

  if (!disp_mode_lock)
    {
      disp_mode++;
      update_disp = true;
    }
}

    
  

  
